﻿using System;
using System.Collections.Generic;
using System.Text;

using System.IO;
using System.IO.Packaging;
using System.Xml;
using System.Xml.Xsl;

namespace testDocX2Html
{
    public static class ConvertDocX2Html
    {
        const string DATE_TIME_FORMAT = "MM-dd-yyyyTHH-mm-ss";
        const string URI_PATH_WORD = "/word";
        const string URI_DOCUMENT_XML = URI_PATH_WORD + "/document.xml";
        const string URI_DOCUMENT_XML_RELS = URI_PATH_WORD + "/_rels/document.xml.rels";
        const string URI_SCHEMA_OPENXMLFORMATS_ORG = "http://schemas.openxmlformats.org";
        const string URI_SCHEMA_OFFICEDOCUMENT_RELATIONSHIP_IMAGE = URI_SCHEMA_OPENXMLFORMATS_ORG 
            + "/officeDocument/2006/relationships/image";
        const string XPATH_SELECT_RELATIONSHIP_IMAGES = "child::node()[local-name()='Relationship' and @Type='" 
            + URI_SCHEMA_OFFICEDOCUMENT_RELATIONSHIP_IMAGE + "']";
        const string XML_DOCIMAGE_INIT_DATA = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\" ?>\r\n<Relationships/>";

        public static bool GetHtmlData(
            string fileDocx, 
            string fileXslt,
            bool extractImages,
            out string outputFolder,
            out string htmlContents, 
            out string error)
        {
            bool bRetVal = false;
            error = htmlContents = String.Empty;
            //outputFolder = String.Empty;

            //////////////////////////////////////////////////////////
            // Get docx content as xml document & list of images
            XmlDocument xmlDocx = null;
            XmlDocument xmlDocImages = null; 
            ConvertDocX2Html.GetDocxData(
                fileDocx,
                extractImages,
				out outputFolder,
				out xmlDocx,
                out xmlDocImages, 
                out error);
            if (!String.IsNullOrEmpty(error))
            {
                return bRetVal;
            }

            //////////////////////////////////////////////////////////
            // Transform xml document into html-string
            try
            {
                if (xmlDocx == null || xmlDocx.DocumentElement == null)
                {
                    throw new NullReferenceException("xmlDocx.DocumentElement");
                }
                XsltArgumentList xsltArgumentList = null;
                if (xmlDocImages != null && xmlDocImages.DocumentElement != null 
                    && xmlDocImages.DocumentElement.ChildNodes.Count > 0)
                {
                    xsltArgumentList = new XsltArgumentList();
                    xsltArgumentList.AddParam("prmListImages", 
                        xmlDocImages.DocumentElement.NamespaceURI, 
                        xmlDocImages.DocumentElement.CreateNavigator()); 
                }
                XslCompiledTransform xslXsltDocTrans = new XslCompiledTransform(/*true*/);
                xslXsltDocTrans.Load(fileXslt);
                Encoding enc = Encoding.UTF8;
                using (MemoryStream msHtmlDoc = new MemoryStream())
                {
                    using (XmlTextWriter xtwXsltDoc = new XmlTextWriter(msHtmlDoc, enc))
                    {
                        xslXsltDocTrans.Transform(xmlDocx.CreateNavigator(), xsltArgumentList, xtwXsltDoc);
                        htmlContents = enc.GetString(msHtmlDoc.GetBuffer(), 0, (int)msHtmlDoc.Length);
                        xtwXsltDoc.Close();
                    }
                    msHtmlDoc.Close();
                    bRetVal = !String.IsNullOrEmpty(htmlContents);
                }
            }
            catch (Exception ex)
            {
                error = ex.ToString(); 
            }
            return bRetVal;
        }

        /////////////////////////////////////////////////////////////
        // Get docx content as xml document & list of images
        public static void GetDocxData(
            string fileDocx,
            bool extractImages,
			out string outputFolder,
			out XmlDocument xmlDocx,
            out XmlDocument xmlDocImages,
            out string error)
        {
            xmlDocx = null;
            xmlDocImages = null;
			error = outputFolder = String.Empty;
            try
            {
                if (!File.Exists(fileDocx))
                {
                    throw new FileNotFoundException(fileDocx); 
                }
                using (Package package =
                    Package.Open(fileDocx, FileMode.Open, FileAccess.Read))
                {
                    if (package == null)
                    {
                        throw new NullReferenceException("Package for " + fileDocx); 
                    }

                    ////////////////////////////////////////////
                    // Get xml-document
                    PackagePart packagePart = package.GetPart(new Uri(URI_DOCUMENT_XML, UriKind.Relative));
                    if (packagePart == null)
                    {
                        throw new NullReferenceException("PackagePart for " 
                            + URI_DOCUMENT_XML);
                    }
                    using (Stream ppStrmXmlDoc = packagePart.GetStream(FileMode.Open, FileAccess.Read))
                    {
                        if (ppStrmXmlDoc == null)
                        {
                            throw new NullReferenceException("PackagePart.GetStream() for " 
                                + URI_DOCUMENT_XML);
                        }
                        byte[] buffer = new byte[ppStrmXmlDoc.Length];
                        ppStrmXmlDoc.Seek(0, SeekOrigin.Begin);
                        ppStrmXmlDoc.Read(buffer, 0, (int)ppStrmXmlDoc.Length);
                        xmlDocx = new XmlDocument();
                        xmlDocx.LoadXml(Encoding.UTF8.GetString(buffer, 0, buffer.Length));    
                        ppStrmXmlDoc.Close();
                    }

                    // Use list of images?
					if (extractImages)
                    {
                        ////////////////////////////////////////////
                        // Get list of images
                        string relsNamespaceURI = String.Empty;  
                        packagePart = package.GetPart(new Uri(URI_DOCUMENT_XML_RELS, UriKind.Relative));
                        if (packagePart != null)
                        {
                            using (Stream ppStrmXmlDoc = packagePart.GetStream(FileMode.Open, 
                                FileAccess.Read))
                            {
                                if (ppStrmXmlDoc == null)
                                {
                                    throw new NullReferenceException("PackagePart.GetStream() for " 
                                        + URI_DOCUMENT_XML_RELS);
                                }
                                byte[] buffer = new byte[ppStrmXmlDoc.Length];
                                ppStrmXmlDoc.Seek(0, SeekOrigin.Begin);
                                ppStrmXmlDoc.Read(buffer, 0, (int)ppStrmXmlDoc.Length);
                                XmlDocument xmlDocRels = new XmlDocument();
                                xmlDocRels.LoadXml(Encoding.UTF8.GetString(buffer, 0, buffer.Length));
                                if (xmlDocRels.DocumentElement != null 
                                    && xmlDocRels.DocumentElement.ChildNodes.Count > 0)
                                {
                                    XmlNamespaceManager nsMgr = new XmlNamespaceManager(xmlDocRels.NameTable);
                                    if (!String.IsNullOrEmpty(xmlDocRels.DocumentElement.NamespaceURI))
                                    {
                                        relsNamespaceURI = xmlDocRels.DocumentElement.NamespaceURI;
                                        nsMgr.AddNamespace("", relsNamespaceURI);
                                    }
                                    XmlNodeList xmlNdLstImages = 
                                        xmlDocRels.DocumentElement.SelectNodes(XPATH_SELECT_RELATIONSHIP_IMAGES, nsMgr);
                                    if (xmlNdLstImages != null && xmlNdLstImages.Count > 0)
                                    {
                                        //////////////////////////////////////////////////////////////
                                        // Copy xmlNdLstImages into new XmlDocument (xmlDocImages)
                                        xmlDocImages = new XmlDocument();
                                        xmlDocImages.LoadXml(XML_DOCIMAGE_INIT_DATA);
                                        XmlNode newXn = null;
                                        XmlAttribute newXa = null;
                                        foreach (XmlNode xnImage in xmlNdLstImages)
                                        {
                                            newXn = xmlDocImages.CreateNode(XmlNodeType.Element, 
                                                xnImage.LocalName, "");
                                            foreach (XmlAttribute xa in xnImage.Attributes)
                                            {
                                                newXa = xmlDocImages.CreateAttribute(xa.LocalName);
                                                newXa.Value = xa.Value;
                                                newXn.Attributes.Append(newXa);
                                            }
                                            xmlDocImages.DocumentElement.AppendChild(newXn);
                                        }
                                    }
                                }
                                ppStrmXmlDoc.Close();
                            }
                        }

                        ////////////////////////////////////////////
                        // Extract images
						if (xmlDocImages != null && xmlDocImages.DocumentElement != null
							&& xmlDocImages.DocumentElement.ChildNodes.Count > 0)
						{
							// Create output folder
							CreateOutputFolder(extractImages, fileDocx,
								out outputFolder, out error);
							if (String.IsNullOrEmpty(error) && !String.IsNullOrEmpty(outputFolder))
							{
								XmlAttribute xmlAttr = null;
								string attrTarget = String.Empty;
								string imageFolder = String.Empty;
								string imageFile = String.Empty;
								foreach (XmlNode xnImage in xmlDocImages.DocumentElement.ChildNodes)
								{
									xmlAttr = xnImage.Attributes["Target"];
									if (xmlAttr == null)
									{
										continue;
									}
									attrTarget = xmlAttr.InnerText;
									if (String.IsNullOrEmpty(attrTarget))
									{
										continue;
									}
									xmlAttr = xnImage.Attributes["TargetMode"];
									if (xmlAttr != null)
									{
										continue;
									}
									if (attrTarget[0] != '/')
									{
										attrTarget = "/" + attrTarget;
									}
									imageFile = outputFolder + attrTarget.Replace('/', '\\');
									attrTarget = URI_PATH_WORD + attrTarget;
									packagePart = package.GetPart(new Uri(attrTarget, UriKind.Relative));
									if (packagePart != null)
									{
										int posLBS = imageFile.LastIndexOf('\\');
										if (posLBS != (-1))
										{
											imageFolder = imageFile.Substring(0, posLBS);
										}
										else
										{
											throw new ArgumentException(imageFolder + " for "
												+ attrTarget);
										}
										if (!Directory.Exists(imageFolder))
										{
											Directory.CreateDirectory(imageFolder);
											if (!Directory.Exists(imageFolder))
											{
												throw new DirectoryNotFoundException(imageFolder + " for "
													+ attrTarget);
											}
										}
										using (Stream ppStrmXmlDoc = packagePart.GetStream(FileMode.Open,
											FileAccess.Read))
										{
											if (ppStrmXmlDoc == null)
											{
												throw new NullReferenceException("PackagePart.GetStream() for "
													+ attrTarget);
											}
											byte[] buffer = new byte[ppStrmXmlDoc.Length];
											ppStrmXmlDoc.Seek(0, SeekOrigin.Begin);
											ppStrmXmlDoc.Read(buffer, 0, (int)ppStrmXmlDoc.Length);
											if (File.Exists(imageFile))
											{
												File.Delete(imageFile);
											}
											using (FileStream fs = new FileStream(imageFile, FileMode.CreateNew))
											{
												if (fs == null)
												{
													throw new NullReferenceException("FileStream for "
														+ imageFile);
												}
												fs.Write(buffer, 0, (int)ppStrmXmlDoc.Length);
												fs.Flush();
												fs.Close();
											}
											ppStrmXmlDoc.Close();
										}
									}
								}
							}
						}
                    }
					else
					{
						// Create output folder without images
						CreateOutputFolder(false, fileDocx,
							out outputFolder, out error);
					}

                    // Close package for docx-file
                    package.Close(); 
                }
            }
            catch (Exception ex)
            {
                error = ex.ToString(); 
            }
        }

		//////////////////////////////////////////////////////////
		// Create output folder
		public static void CreateOutputFolder(bool extractImages, string fileDocx, 
			out string outputFolder, out string error)
		{
			outputFolder = error = String.Empty;  
			try
			{
				if (String.IsNullOrEmpty(fileDocx))
				{
					throw new ArgumentNullException("fileDocx");
				}
				int nPos = (-1);
				if (!extractImages)
				{
					if ((nPos = fileDocx.LastIndexOf('\\')) != (-1))
					{
						outputFolder = fileDocx.Substring(0, nPos);
					}
				}
				else
				{
					outputFolder = fileDocx;
					outputFolder += "_" + DateTime.Now.ToString(DATE_TIME_FORMAT);
				}
				if (!String.IsNullOrEmpty(outputFolder) && !Directory.Exists(outputFolder))
				{
					Directory.CreateDirectory(outputFolder);
					if (!Directory.Exists(outputFolder))
					{
						throw new DirectoryNotFoundException(outputFolder);
					}
				}
			}
			catch (Exception ex)
			{
				error = ex.ToString();
			}
		}
    }
}
